import jquery.*
import org.w3c.dom.*
import kotlin.browser.document
import kotlin.browser.window

val canvas = initalizeCanvas()
val vis = VisualME()

fun main(args: Array<String>) {
    jq {
        vis.setNumberOfBranches(15); //15
        vis.setFlexibility(20); //20
        vis.setScreenProportions(4);
        vis.setSpeed(1);
        
        vis.setVisualizationAutochange(true);
        vis.initialize()
    }
    jq(canvas).mousedown {
        var offsetX = 0.toDouble()
        var offsetY = 0.toDouble()
        var element: HTMLElement? = canvas
        while (element != null) {
            val el: HTMLElement = element
            offsetX += el.offsetLeft
            offsetY += el.offsetTop
            element = el.offsetParent as HTMLElement?
        }
        vis.onTouch(it.pageX - offsetX, it.pageY - offsetY)
    }
}

fun initalizeCanvas(): HTMLCanvasElement {
    val canvas = document.createElement("canvas") as HTMLCanvasElement
    val context = canvas.getContext("2d") as CanvasRenderingContext2D
    context.canvas.width  = window.innerWidth.toInt();
    context.canvas.height = window.innerHeight.toInt();
    context.fillStyle = "rgba(0,0,0,0.1)"
    document.body!!.appendChild(canvas)
    return canvas
}

class VisualME() {
    val context = canvas.getContext("2d") as CanvasRenderingContext2D
    val height = canvas.height
    val width = canvas.width
    
    private val MAX_DELTA_COLOR = 20

    // Основные параметры визуализации
    private var numberOfBranches:Int = 0
    private var flexibility:Int = 0
    private var screenProportions:Double = 0.toDouble()
    private var speed:Int = 0
    // Параметры состояний
    private var shapeAutochange:Boolean = false
    private var visualizationAutochange:Boolean = false
    // Состояния изменения компонент цвета
    private var redChange:Boolean = false
    private var greenChange:Boolean = false
    private var blueChange:Boolean = false
    private var rc:Int = 0
    private var gc:Int = 0
    private var bc:Int = 0
    private var rgbChangeMode:Int = 0
    private var visualMode:Int = 0
    private var colorMode:Int = 0
    private var startAngleInt:Int = 0
    private var angleDelta:Double = 0.toDouble()// Угол между ветвями
    private var len:Double = 0.toDouble() // Длина одного звена
    private var angles:FloatArray = FloatArray(0)
    
    fun randomNextInt(max: Int) = (Math.random() * max).toInt()
        
    fun initialize() {
        // Цвета (переходы)
        rc = 0
        gc = 120
        bc = 0
        // Режим изменения цветов (переходов)
        rgbChangeMode = 1
        // Вид визуализации
        visualMode = 1
        // Режим переходов цветов
        colorMode = randomNextInt(MAX_DELTA_COLOR) + 1
        // Начальный угол поворота ветви
        startAngleInt = randomNextInt(360)
        // Вычисляем необходимые параметры
        if (width > height) {
            len = height / screenProportions / flexibility
        } else {
            len = width / screenProportions / flexibility
        }
        window.setInterval({ paint() }, 60);
        window.setInterval({ threadUpdate() }, 2000);
        window.setInterval({ update() }, 50);
    }
    
    /**
     * Установка гибкости ветвей.
     * @param flexibility значение гибкости
     */
    fun setFlexibility(flexibility:Int) {
        this.flexibility = flexibility
        angles = FloatArray(flexibility)
    }
    
    /**
     * Установка количества ветвей.
     * @param numberOfBranches
     */
    fun setNumberOfBranches(numberOfBranches:Int) {
        this.numberOfBranches = numberOfBranches
        // Угол между ветвями
        angleDelta = 2 * Math.PI / numberOfBranches
    }
    
    /**
     * Установка пропорций экрана и визуализации.
     * @param screenProportions
     */
    fun setScreenProportions(screenProportions:Int) {
        this.screenProportions = 4 / screenProportions.toDouble()
    }
    
    /**
     * Установка скорости анимации.
     * @param speed
     */
    fun setSpeed(speed:Int) {
        this.speed = 10 - speed
    }
    
    /**
     * Установка автосмены визуализации.
     * @param visualizationAutochange
     */
    fun setVisualizationAutochange(visualizationAutochange:Boolean) {
        this.visualizationAutochange = visualizationAutochange
    }
    
    fun paint() {
        val paint = context
        paint.fillRect(0.0, 0.0, width.toDouble(), height.toDouble())        
        
        paint.save()
        if (randomNextInt(20) == 5) startAngleInt = randomNextInt(360)
        // Интерполяция углов между щупальцами
        val startAngle = startAngleInt * Math.PI / 180
        angles[0] = (angles[0] + Math.sin(startAngle).toFloat() / speed.toFloat())
        for (i in 1..flexibility - 1) {
            angles[i] = (angles[i] + (angles[i - 1] - angles[i]) * 0.1.toFloat())
        }
        
        for (j in 0..numberOfBranches - 1) {
            var x = 0.5f * width
            var y = 0.5f * height
            var tx = 0f
            var ty = 0f
            val temp1 = j * angleDelta + angles[1]
            for (i in 1..flexibility - 1) {
                val temp2 = j * angleDelta + angles[i]
                when (visualMode) {
                    1 -> {
                        tx = (x + Math.cos(temp1) * len) as Float
                        ty = (y + Math.sin(temp2) * len) as Float
                        x = x + (Math.cos(temp2) * len) as Float
                    }
                    2 -> {
                        tx = (x + Math.sin(temp2) * len) as Float
                        ty = (y + Math.cos(temp2) * len) as Float
                    }
                    3 -> {
                        tx = (x + Math.tan(temp1) * len) as Float
                        ty = (y + Math.cos(temp2) * len) as Float
                        y = y - (Math.sin(temp2) * len) as Float
                    }
                    4 -> {
                        tx = (x + Math.tan(temp1) * len) as Float
                        ty = (y + Math.cos(temp2) * len) as Float
                        x = y - (Math.sin(temp2) * len) as Float
                    }
                    5 -> {
                        tx = (x + Math.tan(temp1) * len) as Float
                        ty = (y + Math.cos(temp2) * len) as Float
                        x = (x * ty * angleDelta + angles[i] - (Math.sin(temp2) * len)) as Float
                    }
                    6 -> {
                        x = x + (Math.sin(temp2) * len) as Float
                        y = y + (Math.cos(temp2) * len) as Float
                        tx = (x + Math.cos(temp1) * len) as Float
                        ty = (y + Math.sin(temp1) * len) as Float
                        x = x + (Math.sin(temp2) * len) as Float
                        y = y + (Math.cos(temp2) * len) as Float
                    }
                }
                
                if (colorMode == 1) setColor(paint, rc, gc, bc)
                else {
                    val deltaColor = 255 - 255 * i / flexibility
                    when (colorMode) {
                        2 -> setColor(paint, deltaColor, 0, 255)
                        3 -> setColor(paint, deltaColor, 255, 0)
                        4 -> setColor(paint, deltaColor, 255, 255)
                        5 -> setColor(paint, 0, deltaColor, 255)
                        6 -> setColor(paint, 255, deltaColor, 0)
                        7 -> setColor(paint, 255, deltaColor, 255)
                        8 -> setColor(paint, 0, 255, deltaColor)
                        9 -> setColor(paint, 255, 0, deltaColor)
                        10 -> setColor(paint, 255, 255, deltaColor)
                        11 -> setColor(paint, deltaColor, deltaColor, 0)
                        12 -> setColor(paint, deltaColor, deltaColor, 255)
                        13 -> setColor(paint, deltaColor, 0, deltaColor)
                        14 -> setColor(paint, deltaColor, 255, deltaColor)
                        15 -> setColor(paint, 0, deltaColor, deltaColor)
                        16 -> setColor(paint, 255, deltaColor, deltaColor)
                        17 -> setColor(paint, rc, deltaColor, 255)
                        18 -> setColor(paint, 255, gc, deltaColor)
                        19 -> setColor(paint, 255, deltaColor, bc)
                        20 -> setColor(paint, bc, rc, deltaColor)
                    }
                }
                canvas.drawLine(x, y, tx, ty, paint)

                x = tx
                y = ty
            }
        }
        paint.restore()
    }
    
    fun onTouch(x:Double, y:Double) {
        if (y >= (height / 2)) {
            colorMode++
            if (colorMode > MAX_DELTA_COLOR) {
                colorMode = 1
            }
        } else {
            visualMode++
            if (visualMode > 6) {
                visualMode = 1
            }
        }
    }
    
    fun update() {
        rgb()
    }
    
    fun threadUpdate() {
        val i = randomNextInt(20)
        if (visualizationAutochange) {
            if (i == 16)
                colorMode = randomNextInt(MAX_DELTA_COLOR) + 1
            else if (i == 19) visualMode = randomNextInt(6) + 1
        }
    }
    
    private fun setColor(context:CanvasRenderingContext2D, red:Int, green:Int, blue:Int) {
        context.strokeStyle = "rgb($red,$green,$blue)";
    }
    
    fun HTMLCanvasElement.drawLine(x1:Float, y1:Float, x2:Float,y2:Float, paint:CanvasRenderingContext2D) {
        paint.beginPath()
        paint.moveTo(x1.toDouble(), y1.toDouble())
        paint.lineTo(x2.toDouble(), y2.toDouble())
        paint.stroke()
        paint.closePath()
    }
    
    private fun rgb() {
        if (rgbChangeMode == 1) {
            rc++;
            if (rc > 253) {
                rc--;
                redChange = true;
            }
            if (redChange == true) {
                rc = rc - 2;
            }
            if (rc < 18 && redChange == true) {
                rc = 0;
                redChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 2) {
            gc++;
            if (gc > 253) {
                gc--;
                greenChange = true;
            }
            if (greenChange == true) {
                gc = gc - 2;
            }
            if (gc < 18 && greenChange == true) {
                gc = 0;
                greenChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 3) {
            bc++;
            if (bc > 253) {
                bc--;
                blueChange = true;
            }
            if (blueChange == true) {
                bc = bc - 2;
            }
            if (bc < 18 && blueChange == true) {
                bc = 0;
                blueChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 4) {
            rc++;
            gc++;
            if (rc > 253) {
                rc--;
                gc--;
                redChange = true;
            }
            if (redChange == true) {
                rc = rc - 2;
                gc = gc - 2;
            }
            if (rc < 18 && redChange == true) {
                rc = 0;
                gc = 0;
                redChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 5) {
            rc++;
            bc++;
            if (rc > 253) {
                rc--;
                bc--;
                redChange = true;
            }
            if (redChange == true) {
                rc = rc - 2;
                bc = bc - 2;
            }
            if (rc < 18 && redChange == true) {
                rc = 0;
                bc = 0;
                redChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 6) {
            gc++;
            bc++;
            if (gc > 253) {
                gc--;
                bc--;
                greenChange = true;
            }
            if (greenChange == true) {
                gc = gc - 2;
                bc = bc - 2;
            }
            if (gc < 18 && greenChange == true) {
                gc = 0;
                bc = 0;
                greenChange = false;
                rgbChangeMode++;
            }
        }
        else if (rgbChangeMode == 7) {
            rc++;
            gc++;
            bc++;
            if (gc > 253) {
                rc--;
                gc--;
                bc--;
                greenChange = true;
            }
            if (greenChange == true) {
                rc = rc - 2;
                gc = gc - 2;
                bc = bc - 2;
            }
            if (gc < 18 && greenChange == true) {
                rc = 0;
                gc = 0;
                bc = 0;
                greenChange = false;
                rgbChangeMode = 1;
            }
        }
    }
}